#![cfg(target_arch = "x86_64")]
use x86_64::structures::idt::{InterruptDescriptorTable, InterruptStackFrame, PageFaultErrorCode};
use crate::{try_print, try_println};
use core::arch::asm;
use spin::Mutex;
use crate::timer;
use pic8259::ChainedPics;

pub const PIC_1_OFFSET: u8 = 32;
pub const PIC_2_OFFSET: u8 = PIC_1_OFFSET + 8;

pub static PICS: Mutex<ChainedPics> =
    Mutex::new(unsafe { ChainedPics::new(PIC_1_OFFSET, PIC_2_OFFSET) });

static IDT: Mutex<InterruptDescriptorTable> = Mutex::new(InterruptDescriptorTable::new());

#[derive(Debug, Clone, Copy)]
#[repr(u8)]
pub enum IRQ {
    Timer = PIC_1_OFFSET,
    Keyboard = PIC_1_OFFSET + 1,
}

extern "x86-interrupt" fn breakpoint_handler(stack_frame: InterruptStackFrame)
{
    try_println!("Breakpoint!\n{:#?}", stack_frame);
}

extern "x86-interrupt" fn page_fault_handler(stack_frame: InterruptStackFrame, faulting_address: PageFaultErrorCode)
{
    try_println!("Page fault at {:#?}\n{:#?}", faulting_address, stack_frame);
}

extern "x86-interrupt" fn double_fault_handler(stack_frame: InterruptStackFrame, _error_code: u64) -> ! {
    panic!("Double fault!\n{:#?}", stack_frame);
}

extern "x86-interrupt" fn timer_handler(_stack_frame: InterruptStackFrame)
{
    timer::tick();
    unsafe { PICS.lock().notify_end_of_interrupt(IRQ::Timer as u8); }
}

extern "x86-interrupt" fn keyboard_handler(_stack_frame: InterruptStackFrame)
{
    use pc_keyboard::{layouts, DecodedKey, HandleControl, Keyboard, ScancodeSet1};

    use x86_64::instructions::port::Port;
    static KEYBOARD: Mutex<Keyboard<layouts::Us104Key, ScancodeSet1>> = Mutex::new(Keyboard::new(HandleControl::Ignore));

    let mut keyboard = KEYBOARD.lock();
    let mut port = Port::new(0x60);

    let scancode: u8 = unsafe { port.read() };
    if let Ok(Some(key_event)) = keyboard.add_byte(scancode) {
        if let Some(key) = keyboard.process_keyevent(key_event) {
            match key {
                DecodedKey::Unicode(character) => try_print!("{}", character),
                DecodedKey::RawKey(key) => try_print!("{:?}", key),
            }
        }
    }

    unsafe {
        PICS.lock()
            .notify_end_of_interrupt(IRQ::Keyboard as u8);
    }
}

pub fn load()
{
    let mut idt = IDT.lock();
    idt.breakpoint.set_handler_fn(breakpoint_handler);
    idt.page_fault.set_handler_fn(page_fault_handler);
    idt.double_fault.set_handler_fn(double_fault_handler);
    idt[IRQ::Timer as usize].set_handler_fn(timer_handler);
    idt[IRQ::Keyboard as usize].set_handler_fn(keyboard_handler);

    unsafe { 
        idt.load_unsafe(); 
    }

    unsafe { 
        PICS.lock().initialize();
        PICS.lock().write_masks(0b11111100, 0b11111111)
    }
}

pub fn enable()
{
    unsafe { asm!("sti"); }
}

pub fn disable()
{
    unsafe { asm!("cli"); }
}