#![allow(dead_code)]
use crate::println;

use spin::RwLock;

static mut TIMER: u64 = 0; // writes to aligned 64-bit integers are atomic, at least on x86_64
static BOOT_TIMESTAMP: RwLock<i64> = RwLock::new(0);

use crate::arch::timer;

pub fn init(boot_timestamp: i64)
{
    *BOOT_TIMESTAMP.write() = boot_timestamp;
    timer::platform_init();
}

pub fn tick()
{
    unsafe { TIMER += 1 }
}

pub fn count() -> u64 {
    unsafe { TIMER }
}

const MS_PER_SECOND: u64 = 1000;
const US_PER_SECOND: u64 = 1000000;
const NS_PER_SECOND: u64 = 1000000000;

pub fn second_count() -> u64 {
    return count() / timer::frequency();
}

// FIXME: Propagate overflows. A system call may want to return EOVERFLOW instead of panicking.
pub fn millisecond_count() -> u64 {
    return count().checked_mul(MS_PER_SECOND).unwrap() / timer::frequency();
}

pub fn microsecond_count() -> u64 {
    return count().checked_mul(US_PER_SECOND).unwrap() / timer::frequency();
}

pub fn nanosecond_count() -> u64 {
    return count().checked_mul(NS_PER_SECOND).unwrap() / timer::frequency();
}

pub fn boot_timestamp() -> i64 {
    *BOOT_TIMESTAMP.read()
}

// FIXME: These functions are verbose.
pub fn second_time() -> i64 {
    i64::try_from(second_count()).unwrap().checked_add(boot_timestamp()).unwrap()
}

pub fn millisecond_time() -> i64 {
    i64::try_from(millisecond_count()).unwrap().checked_add(boot_timestamp().checked_mul(MS_PER_SECOND as i64).unwrap()).unwrap()
}

pub fn microsecond_time() -> i64 {
    i64::try_from(microsecond_count()).unwrap().checked_add(boot_timestamp().checked_mul(US_PER_SECOND as i64).unwrap()).unwrap()
}

pub fn nanosecond_time() -> i64 {
    i64::try_from(nanosecond_count()).unwrap().checked_add(boot_timestamp().checked_mul(NS_PER_SECOND as i64).unwrap()).unwrap()
}

pub fn isleap(year: i32) -> bool
{
    year % 4 == 0 && (year % 100 != 0 || year % 400 == 0)
}

fn make_yday(year: i32, month: i32) -> i32 {
    let upto: [i16; 12usize] = [0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334];
    let mut yd: i16;

    yd = upto[(month - 1) as usize];
    if month > 2 && isleap(year)
    { 
        yd += 1;
    }
    yd as i32
}

pub fn broken_down_to_unix(year: i64, month: i64, day: i64, hour: i64, min: i64, sec: i64) -> i64 {
    return sec + min * 60 + hour * 3600 + (make_yday(year as i32, month as i32) + (day - 1) as i32) as i64 * 86400 + (year - 70) * 31536000 + ((year - 69) / 4) * 86400 -
           ((year - 1) / 100) * 86400 + ((year + 299) / 400) * 86400;
}

fn bcd_number_to_decimal(num: u8) -> i32
{
    return (((num >> 4) * 10) + (num & 0xf)) as i32;
}

pub fn decode_timestamp(datetime: [u8; 8]) -> i64 {
    let year = (bcd_number_to_decimal(datetime[0]) * 100) + bcd_number_to_decimal(datetime[1]);
    let month = bcd_number_to_decimal(datetime[2]);
    let day = bcd_number_to_decimal(datetime[3]);
    let hour = bcd_number_to_decimal(datetime[4]);
    let min = bcd_number_to_decimal(datetime[5]);
    let sec = bcd_number_to_decimal(datetime[6]);
    println!("{}-{}-{} {}:{}:{}", year, month, day, hour, min, sec);
    broken_down_to_unix((year - 1900) as i64, month as i64, day as i64, hour as i64, min as i64, sec as i64)
}

use crate::bootboot::BOOTBOOT;
use crate::util::get_bootboot;

pub fn decode_boot_timestamp() -> i64 {
    let boot: &BOOTBOOT = get_bootboot();
    decode_timestamp(boot.datetime)
}